import asyncio
import datetime
import logging

from pscr_common.utils import get_utcnow_with_tzinfo, chunker
from sqlalchemy import select, bindparam
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.types import TIMESTAMP

from feed_manager.sql.models import OrganizationWatchlistState, Watchlist
from feed_manager.utils import aws_client
from feed_manager.utils.config import Config
from pscr_common.utils import datetime_to_ms_timestamp

_logger = logging.getLogger(__name__)


class OrgWatchlistStateUpdater(object):
    def __init__(self, config: Config, now: datetime.datetime = None):
        self.config = config
        self.s3_bucket = config.get("watchlist", "augmentation_bucket")
        self.dirty_timestamp = now if now else get_utcnow_with_tzinfo()
        self.dirty_tag = f"dirty={datetime_to_ms_timestamp(self.dirty_timestamp)}"
        self.insert_stmt = None

    def set_dirty_for_org_id(self, org_id: str) -> "OrgWatchlistStateUpdater":
        self.insert_stmt = (
            insert(OrganizationWatchlistState)
            .values(org_id=org_id, dirty_timestamp=self.dirty_timestamp)
            .returning(OrganizationWatchlistState.org_id)
        )

        return self

    def set_dirty_for_watchlist_id_select(
        self, sel_stmt, org_id: str = None
    ) -> "OrgWatchlistStateUpdater":
        wl_sel_stmt = (
            select(
                [
                    Watchlist.org_id,
                    bindparam(
                        "dirty_timestamp",
                        self.dirty_timestamp,
                        type_=TIMESTAMP(timezone=True),
                    ),
                ]
            )
            .where(Watchlist.id.in_(sel_stmt))
            .distinct(Watchlist.org_id)
        )

        if org_id:
            wl_sel_stmt = wl_sel_stmt.where(Watchlist.org_id == org_id)

        self.insert_stmt = (
            insert(OrganizationWatchlistState)
            .from_select(
                [
                    OrganizationWatchlistState.org_id,
                    OrganizationWatchlistState.dirty_timestamp,
                ],
                wl_sel_stmt,
            )
            .returning(OrganizationWatchlistState.org_id)
        )

        return self

    async def execute(self, conn):
        upsert_stmt = self.insert_stmt.on_conflict_do_nothing(
            index_elements=[OrganizationWatchlistState.org_id]
        )

        await asyncio.gather(
            *(
                self.set_s3_dirty_tag(chunk)
                for chunk in chunker(
                    50, (res["org_id"] for res in await conn.execute(upsert_stmt))
                )
            )
        )

    async def create_dirty_object(self, s3_client, org_id, match_type):
        key = self._dirty_key(org_id, match_type)
        _logger.info(
            "Create dirty bucket:%s key:%s tagging:%s",
            self.s3_bucket,
            key,
            self.dirty_tag,
        )
        return await s3_client.put_object(
            Bucket=self.s3_bucket,
            Key=key,
            Tagging=self.dirty_tag,
            Body=b"",
        )

    async def set_s3_dirty_tag(self, org_ids: [str]) -> None:
        async with aws_client(self.config, "s3") as s3_client:
            await asyncio.gather(
                *(
                    self.create_dirty_object(s3_client, org_id, "ingress")
                    for org_id in org_ids
                ),
                *(
                    self.create_dirty_object(s3_client, org_id, "query")
                    for org_id in org_ids
                ),
            )

    async def delete_s3_dirty_tag(
        self, org_id: str, match_type: str, s3_client
    ) -> None:
        _logger.info(
            "Deleting dirty tag (dirty=0) orgId:%s match_type:%s", org_id, match_type
        )
        await s3_client.put_object_tagging(
            Bucket=self.s3_bucket,
            Key=self._dirty_key(org_id, match_type),
            Tagging={"TagSet": [{"Key": "dirty", "Value": "0"}]},
        )

    def _dirty_key(self, org_id: str, match_type: str):
        return f"watchlists/org={org_id}/watchlist.{match_type}.json.dirty"
