import os
import logging
import uuid
from feed_manager.feed_sync.feedclient import FeedClient
from feed_manager.utils.config import Config
from pscr_common.log import setup_logging, clear_logging_handlers

CONFIG_SECTION = "feedsync"
CDC_PUBLIC_FEEDS = (
    "Carbon Black SANS",
    "Carbon Black Community",
    "US Cybercom Malware Alert",
    "Carbon Black Early Access Indicators",
    "Carbon Black Advanced Threats",
    "Carbon Black Endpoint Visibility",
    "Carbon Black TOR",
    "Carbon Black Known IOCs",
    "Carbon Black Endpoint Suspicious Indicators",
    "ATT&CK Framework",
    "Carbon Black Facebook ThreatExchange",
    "Carbon Black AlienVault",
    "AMSI Threat Intelligence",
)

_logger = logging.getLogger(__name__)


def main():
    # read configs
    config = Config.create()

    clear_logging_handlers()
    setup_logging(
        level=config.get("log", "level"),
        svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
        env_id=os.environ.get("ENV_ID", "UNKNOWN"),
        svc_id="TODO",
    )

    _job_run_id = uuid.uuid4()
    _logger.info("Started feed_sync job with id %s", str(_job_run_id))

    # init clients
    target_client = FeedClient(
        domain=config.get(CONFIG_SECTION, "target_domain"),
        prefix=config.get(CONFIG_SECTION, "target_url_root"),
        org_key=config.get(CONFIG_SECTION, "target_org_key"),
        api_token=None,
        add_context=True,
    )
    source_client = FeedClient(
        domain=config.get(CONFIG_SECTION, "source_domain"),
        prefix=config.get(CONFIG_SECTION, "source_url_root"),
        org_key=config.get(CONFIG_SECTION, "source_org_key"),
        api_token=config.get(CONFIG_SECTION, "source_api_token"),
        add_context=False,
    )

    # fetch current feeds from source and target
    target_feed_infos = target_client.get_feeds(include_public=True)
    target_feed_infos = filter_feeds(target_feed_infos)
    source_feed_infos = source_client.get_feeds(include_public=True)
    source_feed_infos = filter_feeds(source_feed_infos)

    _logger.debug(
        "Got %d feeds from source, %d feeds from target",
        len(source_feed_infos),
        len(target_feed_infos),
    )

    # get intersections
    added_feed_infos, deleted_feed_infos, updated_feed_infos = intersect_feeds(
        source_feed_infos, target_feed_infos
    )
    _logger.info(
        "Calculated feeds: %d NEW, %d DELETED, %d UPDATED",
        len(added_feed_infos),
        len(deleted_feed_infos),
        len(updated_feed_infos),
    )

    if added_feed_infos:
        handle_new_feeds(added_feed_infos, source_client, target_client)
    if deleted_feed_infos:
        handle_deleted_feeds(deleted_feed_infos, target_client)
    if updated_feed_infos:
        handle_updated_feeds(
            updated_feed_infos, target_feed_infos, source_client, target_client
        )

    _logger.info("Finished feed_sync job with id %s", _job_run_id)


def filter_feeds(feeds):
    return list(
        filter(
            lambda feed: feed.access == "public" and feed.name in CDC_PUBLIC_FEEDS,
            feeds,
        )
    )


def intersect_feeds(source_feeds, target_feeds):
    source_feeds_set = set(source_feeds)
    target_feeds_set = set(target_feeds)
    added = list(source_feeds_set - target_feeds_set)
    deleted = list(target_feeds_set - source_feeds_set)
    # the resulting set needs to be of elements from the source set because th
    # ID will be used for API calls to the SOURCE
    updated = [item for item in source_feeds_set if item in target_feeds_set]
    return added, deleted, updated


def reset_reports(reports):
    return [
        report._replace(timestamp=int(report.timestamp.timestamp()))
        for report in reports
    ]


def handle_new_feeds(new_feed_infos, source_client, target_client):
    _logger.debug("Entering handle_new_feeds")
    for new_feed_info in new_feed_infos:
        _logger.debug(
            "Handling feed {id: %s, name: %s}", new_feed_info.id, new_feed_info.name
        )

        feed = source_client.get_feed(new_feed_info.id)
        feed = feed._replace(reports=reset_reports(feed.reports))

        target_client.create_feed(feed=feed, public=True)
    _logger.debug("Exiting handle_new_feeds")


def handle_deleted_feeds(deleted_feed_infos, target_client):
    _logger.debug("Entering handle_deleted_feeds")
    for feed_info in deleted_feed_infos:
        _logger.debug("Handling feed {id: %s, name: %s}", feed_info.id, feed_info.name)
        target_client.delete_feed(feed_info.id)
    _logger.debug("Exiting handle_deleted_feeds")


def handle_updated_feeds(
    updated_feed_infos, target_feed_infos, source_client, target_client
):
    _logger.debug("Entering handle_updated_feeds")
    for updated_feed_info in updated_feed_infos:
        _logger.debug(
            "Handling feed {id: %s, name: %s}",
            updated_feed_info.id,
            updated_feed_info.name,
        )
        # updated_feed_infos should have the id-s from the source
        feed = source_client.get_feed(updated_feed_info.id)
        # fetch id from target feed_info to do the update
        target_feed_id = next(
            feed_info.id
            for feed_info in target_feed_infos
            if feed_info == updated_feed_info
        )

        feed_info = feed.feedinfo._replace(id=target_feed_id)
        feed = feed._replace(feedinfo=feed_info)
        feed = feed._replace(reports=reset_reports(feed.reports))

        target_client.update_feed_info(feed_id=target_feed_id, feed_info=feed.feedinfo)
        # create reports actually replaces all feed's reports
        target_client.create_reports(feed_id=target_feed_id, reports=feed.reports)
    _logger.debug("Exiting handle_updated_feeds")


if __name__ == "__main__":
    main()
