import logging
from typing import Any, Dict
from contextlib import asynccontextmanager

from sqlalchemy.ext.asyncio import create_async_engine, AsyncSession
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine.url import URL

from feed_manager.utils.config import Config
from pscr_common.aws import get_boto3_client

_logger = logging.getLogger(__name__)


def get_password(config: Config, config_section: str):
    password_ssm_key = config.get(config_section, "password_ssm_key")
    return (
        get_password_from_ssm(password_ssm_key)
        if password_ssm_key
        else config.get(config_section, "password")
    )


def get_password_from_ssm(key):
    ssm = get_boto3_client("ssm")
    response = ssm.get_parameter(Name=key, WithDecryption=True)

    return response["Parameter"]["Value"]


def get_orm_connect_config(config: Config, read_only=False) -> Dict[str, Any]:
    if read_only and config.getboolean("db.readonly", "enabled", fallback=False):
        section = "db.readonly"
    else:
        section = "db.readwrite"

    return {
        "host": config.get(section, "host"),
        "port": config.getint(section, "port"),
        "database": config.get(section, "name"),
        "username": config.get(section, "user"),
        "password": get_password(config, section),
        "connect_timeout": config.getint(section, "connect_timeout"),
        "min_pool_size": config.getint(section, "min_pool_size"),
        "max_pool_size": config.getint(section, "max_pool_size"),
        "ssl": config.get(section, "ssl"),
    }


def get_engine(config: Config, read_only=False):
    db_config = get_orm_connect_config(config, read_only=read_only)

    connect_timeout = db_config.pop("connect_timeout", None)
    pool_size = db_config.pop("min_pool_size", None)
    max_overflow = db_config.pop("max_pool_size", None) - pool_size
    ssl = db_config.pop("ssl", None)
    if ssl == "True":
        ssl = True

    return create_async_engine(
        URL.create("postgresql+asyncpg", **db_config),
        connect_args={"timeout": connect_timeout, "ssl": ssl},
        pool_size=pool_size,
        max_overflow=max_overflow,
    )


def scoped_session(config: Config):
    pool = ReadWriteDbPool(config)
    pool.initialize()
    return pool.rw_transaction()


class ReadWriteDbPool(object):
    def __init__(self, config: Config):
        self.config = config
        self.rw_engine = None
        self.ro_engine = None
        self.rw_session_maker = None
        self.ro_session_maker = None

    def initialize(self):
        self.rw_engine = get_engine(self.config, False)
        self.rw_session_maker = sessionmaker(bind=self.rw_engine, class_=AsyncSession)

        if self.config.getboolean("db.readonly", "enabled", fallback=False):
            self.ro_engine = get_engine(self.config, True)
            self.ro_session_maker = sessionmaker(
                bind=self.ro_engine, class_=AsyncSession
            )
        else:
            self.ro_engine = self.rw_engine
            self.ro_session_maker = self.rw_session_maker

    @asynccontextmanager
    async def transaction(self, read_only=False):
        session = self.ro_session_maker() if read_only else self.rw_session_maker()

        try:
            yield session
            await session.commit()
        except Exception as e:
            await session.rollback()
            raise
        finally:
            await session.close()

    def rw_transaction(self):
        return self.transaction(read_only=False)

    def ro_transaction(self):
        return self.transaction(read_only=True)

    async def close(self):
        await self.rw_engine.dispose()
        await self.rw_engine.dispose()


def is_unique_violation_exception(e):
    return e.orig.sqlstate == "23505"


def is_fk_violation_exception(e):
    return e.orig.sqlstate == "23503"
