import aiohttp
import asyncio
import logging
from pscr_common.utils import chunker
from typing import Optional, AnyStr

_logger = logging.getLogger(__name__)


class SearchClient:
    def __init__(self, config, client_session=None):
        if client_session is not None:
            self.url = ""
            self.client_session = client_session
        else:
            self.url = config.get("searchservice", "url")
            self.client_session = aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(
                    total=config.getint("searchservice", "timeout", fallback=60)
                )
            )

    async def close(self):
        await self.client_session.close()

    async def get_first_invalid_query(
        self, org_key: str, queries: [str]
    ) -> Optional[AnyStr]:
        _logger.debug("Validating %d queries for org %s", len(queries), org_key)

        futures = [self.make_request(org_key, chunk) for chunk in chunker(250, queries)]

        for response in await asyncio.gather(*futures):
            if response is not None:
                return response

        return None

    async def make_request(self, org_key: str, chunk: [str]) -> Optional[AnyStr]:
        async with self.client_session.post(
            f"{self.url}/internal/search/v1/orgs/{org_key}/processes/search_validation/bulk",
            json={"queries": chunk},
        ) as resp:
            resp.raise_for_status()

            # if any query resulted in an invalid query or a value search query, reject the whole group
            for message in await resp.json():
                if not message.get("valid") or message.get("value_search_query"):
                    return message.get("query")

        return None
