import logging

from sqlalchemy import select

from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.engine.report_engine import ReportEngine
from feed_manager.sql.models import Feed as DbFeed
from feed_manager.utils.tasks.background_task import BackgroundTask
from feed_manager.utils.config import Config
from feed_manager.utils.db import ReadWriteDbPool
from feed_manager.utils.mem_cache import MemCache

_logger = logging.getLogger(__name__)


class UpdatePublicFeedCacheTask(BackgroundTask):
    def __init__(
        self, config: Config, db_pool: ReadWriteDbPool, feed_engine: FeedEngine
    ):
        super(UpdatePublicFeedCacheTask, self).__init__(
            "update_public_feed_cache_task",
            loop_sec=config.get("feedsync", "feed_cache_update_minutes", fallback=15)
            * 60,
            sleep_first=False,
            execute_once=False,
        )
        self.db_pool = db_pool
        self.feed_engine = feed_engine
        self.cache = MemCache()

    async def execute(self, app) -> None:
        sel_stmt = select([DbFeed]).where(DbFeed.access == ReportEngine.PUBLIC_LABEL)
        feeds = list()
        async with self.db_pool.ro_transaction() as conn:
            async for result in await conn.stream(sel_stmt):
                feeds.append(result[0].name)
        if len(feeds):
            self.cache.set("public_feeds", feeds)
            _logger.info(f"Added public feeds to memory cache:\n{feeds}")
        else:
            _logger.warning(
                f"! Retrieved invalid public feeds from RDS !\nfeeds={feeds}\ncurrent={self.cache.get('public_feeds')}"
            )
