import logging

from aiohttp import web
from pscr_common.web.auth import AuthRouteTableDef

from feed_manager.engine.feed_engine import FeedEngine
from feed_manager.engine.watchlist_engine import WatchlistEngine
from feed_manager.utils.mem_cache import MemCache
from feed_manager.web.route_helpers import make_response_json

_logger = logging.getLogger(__name__)


def build_route_table(
    feed_engine: FeedEngine, watchlist_engine: WatchlistEngine
) -> AuthRouteTableDef:
    routes = AuthRouteTableDef()

    @routes.delete("/v1/orgs/{org_key}/purge")
    async def trigger_deprovisioning(request: web.Request):
        cache = MemCache()
        org_key = request.match_info["org_key"]

        orgs = cache.get("orgs_to_deprovision")
        orgs.append(org_key)
        cache.set("orgs_to_deprovision", orgs)

        return web.json_response(data={"message": "Purge request received"}, status=202)

    @routes.get("/v1/orgs/{org_key}/purge", allow_head=False)
    async def get_deprovisioning_status(request: web.Request):
        org_key = request.match_info["org_key"]
        purge_status = "UNKNOWN"
        additional_info = []
        try:
            feeds = await feed_engine.get_feeds(org_key)
            watchlists = await watchlist_engine.get_watchlists(org_key)
            if feeds:
                additional_info.append(f"Org has {len(feeds)} feeds")
                purge_status = "PROGRESS"
            if watchlists:
                additional_info.append(f"Org has {len(watchlists)} watchlists")
                purge_status = "PROGRESS"
            if not any((feeds, watchlists)):
                additional_info.append("Org has no feeds or watchlists")
                purge_status = "SUCCESS"
        except Exception as e:
            exc_info = f"Exception when calling get_deprovisioning_status endpoint: {e}"
            _logger.error(exc_info)
            additional_info.append(exc_info)
            purge_status = "ERROR"

        last_deprovisioning_error = MemCache().get("deprovisioning_last_error")
        if last_deprovisioning_error:
            additional_info.append(last_deprovisioning_error)
            purge_status = "ERROR"

        return make_response_json(
            {
                "purge_status": purge_status,
                "message": f"The purge status is {purge_status}",
                "additional_info": additional_info if additional_info else None,
            }
        )

    return routes
