import logging
import rapidjson
import time

from aiohttp import web
from aiohttp.web_exceptions import HTTPNotFound, HTTPBadRequest, HTTPInternalServerError

from feed_manager.api.models import Report, WatchlistTelemetryList, IOC_V2
from feed_manager.api.models.watchlist_models import (
    WatchlistV2,
    ReportSeverity,
    ReportIOCIgnoreList,
)
from feed_manager.engine.report_engine import ReportEngine
from feed_manager.engine.watchlist_engine import WatchlistEngine
from feed_manager.web.route_helpers import make_response_json
from feed_manager.web.auth import Resource
from pscr_common.utils import get_utcnow_with_tzinfo, datetime_from_sec_timestamp
from pscr_common.web.auth import AuthRouteTableDef, Action
from pscr_common.web.auditlog import call_audit_log

_logger = logging.getLogger(__name__)


def build_route_table(engine: WatchlistEngine) -> AuthRouteTableDef:
    routes = AuthRouteTableDef()

    def _get_required_param(params, key):
        value = params.get(key)
        if not value:
            raise HTTPBadRequest(reason=f"Must provide '{key}'")
        return value

    @routes.delete(
        "/v3/orgs/{org_key}/watchlists/tag", Resource.watchlists, [Action.delete]
    )
    async def remove_watchlist_tag_status_all(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Disable document tagging for ALL watchlists.
        description: Turn off tagging for ALL watchlists in an org.  This is not retroactive for existing watchlist tags.  Future matches will not trigger event tagging.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
        responses:
            204:
                description: Tagging stopped
        """
        org_key = request.match_info["org_key"]
        watchlists = await engine.get_watchlists(org_id=org_key)

        if len(watchlists) == 0:
            return make_response_json()
        if not await engine.remove_tag_status_all(org_key):
            raise HTTPInternalServerError(
                reason=f"Unexpected error. Could not disable tagging for watchlists"
            )

        await call_audit_log(
            request, description=f"Disabled document tagging for ALL watchlists"
        )
        return make_response_json()

    @routes.delete(
        "/v3/orgs/{org_key}/watchlists/alert", Resource.watchlists, [Action.delete]
    )
    async def remove_watchlist_alert_status_all(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Disable alert generation for ALL watchlists.
        description: Turn off alert generation for ALL watchlists in an org.  This is not retroactive for existing watchlist alerts.  Future matches will not trigger alerts.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
        responses:
            204:
                description: Alerting stopped
        """
        org_key = request.match_info["org_key"]

        if not await engine.remove_alert_status_all(org_key):
            raise HTTPInternalServerError(
                reason=f"Unexpected error. Could not disable alerting for watchlists"
            )

        await call_audit_log(
            request, description=f"Disabled alert generation for ALL watchlists"
        )
        return make_response_json()

    @routes.post("/v3/orgs/{org_key}/watchlists", Resource.watchlists, [Action.create])
    async def create_watchlist(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Create new watchlist
        description: Create a new report or classifier watchlist.  Unique watchlist ID will be generated by the service.  Request must specify `report` or `classifier` but not both.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: body
            name: body
            description: Watchlist payload
            required: true
            schema:
              $ref: '#/definitions/WatchlistV2'
        responses:
            200:
                description: Watchlist created.
                schema:
                  $ref: '#/definitions/WatchlistV2'
            400:
                description: invalid watchlist request.
        """
        now = get_utcnow_with_tzinfo()
        params = await request.json(loads=rapidjson.loads)
        params["id"] = WatchlistEngine.gen_watchlist_id()

        watchlist_request = WatchlistV2.from_dict(params)
        org_key = request.match_info["org_key"]

        watchlist = await engine.add_watchlist(org_key, watchlist_request, now)

        await call_audit_log(
            request,
            description=f"Created new watchlist, '{watchlist.name}' (ID: {watchlist.id})",
        )
        return make_response_json(
            WatchlistEngine.remove_empty_alert_classification_field_v2(watchlist)
        )

    @routes.get(
        "/v3/orgs/{org_key}/watchlists",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_watchlists(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get all watchlists
        description: Retrieve all watchlists owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
        responses:
            200:
                description: Array of watchlists
                schema:
                  properties:
                    results:
                      type: array
                      items:
                        $ref: '#/definitions/WatchlistV2'
        """
        org_key = request.match_info["org_key"]

        watchlists = await engine.get_watchlists(org_id=org_key)

        updated_watchlists = []
        for w in watchlists:
            updated_watchlists.append(
                WatchlistEngine.remove_empty_alert_classification_field_v2(w)
            )

        return make_response_json({"results": updated_watchlists})

    @routes.get(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_watchlist(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get watchlist
        description: Retrieve watchlist with `watchlist_id`
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            200:
                description: Return watchlist
                schema:
                  $ref: '#/definitions/WatchlistV2'
            400:
                description: Unknown watchlist.
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        watchlist = (
            await engine.get_watchlists(org_id=org_key, watchlist_id=watchlist_id)
        )[0]

        return make_response_json(
            WatchlistEngine.remove_empty_alert_classification_field_v2(watchlist)
        )

    @routes.put(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}",
        Resource.watchlists,
        [Action.update],
    )
    async def update_watchlist(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Update watchlist
        description: Update watchlist with `watchlist_id`.  This will update the tags and alert status as well as any reports or classifiers attached to the watchlist.  If a field is missing or null (ie tags_enabled) that field will not be updated.  Cannot update report watchlist with empty `report_ids` list.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
          - in: body
            name: body
            description: Watchlist payload
            required: true
            schema:
              $ref: '#/definitions/WatchlistV2'
        responses:
            200:
                description: Return watchlist
                schema:
                  $ref: '#/definitions/WatchlistV2'
            400:
                description: Unknown watchlist or malformed request.
        """
        now = get_utcnow_with_tzinfo()
        params = await request.json(loads=rapidjson.loads)
        watchlist_id = request.match_info["watchlist_id"]

        if watchlist_id != params.get("id", watchlist_id):
            raise HTTPBadRequest(
                reason="Provided watchlist.id must match watchlist id from URL"
            )

        # Set watchlist_id from URL in the json blob before WatchlistV2 is parsed
        #  Once we move to dataclasses we can change this
        params["id"] = watchlist_id

        watchlist_request = WatchlistV2.from_dict(params)

        org_key = request.match_info["org_key"]

        watchlist = await engine.update_watchlist(org_key, watchlist_request, now)

        await call_audit_log(
            request,
            description=f"Updated watchlist, '{watchlist.name}' (ID: {watchlist.id})",
        )
        return make_response_json(
            WatchlistEngine.remove_empty_alert_classification_field_v2(watchlist)
        )

    @routes.delete(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}",
        Resource.watchlists,
        [Action.delete],
    )
    async def remove_watchlist(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Delete watchlist
        description: Remove watchlist with `watchlist_id`.  Existing hits for this watchlist will remain in the system.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            204:
                description: Watchlist deleted
            400:
                description: Unknown watchlist.
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        watchlist_name = await engine.get_watchlist_name(org_key, watchlist_id)

        if not await engine.remove_watchlist(org_key, watchlist_id):
            raise HTTPNotFound(reason=f"Unknown watchlist: {watchlist_id}")

        await call_audit_log(
            request,
            description=f"Deleted watchlist, '{watchlist_name}' (ID: {watchlist_id})",
        )
        return make_response_json()

    @routes.delete(
        "/v3/orgs/{org_key}/watchlists", Resource.watchlists, [Action.delete]
    )
    async def remove_all_watchlists(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Delete ALL watchlists
        description: Remove ALL of an org's watchlists.  Existing hits for this watchlist will remain in the system.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
        responses:
            204:
                description: Watchlist deleted
        """
        org_key = request.match_info["org_key"]
        watchlists = await engine.get_watchlists(org_id=org_key)

        if len(watchlists) == 0:
            return make_response_json()
        if not await engine.remove_all_watchlists(org_key):
            raise HTTPInternalServerError(
                reason="Unexpected error. Unable to remove watchlists"
            )

        await call_audit_log(request, description=f"Deleted ALL watchlists")
        return make_response_json()

    @routes.get(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}/alert",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_watchlist_alert_status(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get watchlist alert status
        description: Retrieve alert status for watchlist with `watchlist_id`.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            200:
                description: Returns alert status
                schema:
                  type: object
                  properties:
                    alert:
                      type: boolean
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        alert = await engine.get_alert_status(org_key, watchlist_id)

        return make_response_json({"alert": alert})

    @routes.put(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}/alert",
        Resource.watchlists,
        [Action.update],
    )
    async def set_watchlist_alert(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Enable watchlist alerts
        description: Turn on alerts for watchlist with `watchlist_id`.  This is not retroactive for existing watchlist hits.  Future hits will trigger alerts.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            200:
                description: Returns alert status
                schema:
                  type: object
                  properties:
                    alert:
                      type: boolean
            400:
                description: Unknown watchlist
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        if not await engine.set_alert_status(org_key, watchlist_id):
            raise HTTPNotFound(reason=f"Unknown watchlist: {watchlist_id}")

        watchlist_name = await engine.get_watchlist_name(org_key, watchlist_id)

        await call_audit_log(
            request,
            description=f"Enabled alerting for watchlist, '{watchlist_name}' (ID: {watchlist_id})",
        )
        return make_response_json({"alert": True})

    @routes.delete(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}/alert",
        Resource.watchlists,
        [Action.delete],
    )
    async def remove_watchlist_alert_status(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Disable watchlist alerts
        description: Turn off alerts for watchlist with `watchlist_id`.  This is not retroactive for existing watchlist alerts.  Future hits will not trigger alerts.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            204:
                description: Returns alert status
            400:
                description: Unknown watchlist
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        if not await engine.remove_alert_status(org_key, watchlist_id):
            raise HTTPNotFound(reason=f"Unknown watchlist: {watchlist_id}")

        watchlist_name = await engine.get_watchlist_name(org_key, watchlist_id)

        await call_audit_log(
            request,
            description=f"Disabled alerting for watchlist, '{watchlist_name}' (ID: {watchlist_id})",
        )
        return make_response_json()

    @routes.get(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}/tag",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_watchlist_tag_status(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get watchlist tag status
        description: Retrieve tag status for watchlist with `watchlist_id`.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            200:
                description: Returns tag status
                schema:
                  type: object
                  properties:
                    tag:
                      type: boolean
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        alert = await engine.get_tag_status(org_key, watchlist_id)

        return make_response_json({"tag": alert})

    @routes.put(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}/tag",
        Resource.watchlists,
        [Action.update],
    )
    async def set_watchlist_tag(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Enable watchlist tags
        description: Turn on tagging for watchlist with `watchlist_id`.  This is not retroactive for existing watchlist matches.  Future matches will trigger event tagging.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            200:
                description: Returns tag status
                schema:
                  type: object
                  properties:
                    tag:
                      type: boolean
            400:
                description: Unknown watchlist
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        if not await engine.set_tag_status(org_key, watchlist_id):
            raise HTTPNotFound(reason=f"Unknown watchlist: {watchlist_id}")

        watchlist_name = await engine.get_watchlist_name(org_key, watchlist_id)

        await call_audit_log(
            request,
            description=f"Enabled tagging for watchlist, '{watchlist_name}' (ID: {watchlist_id})",
        )
        return make_response_json({"tag": True})

    @routes.delete(
        "/v3/orgs/{org_key}/watchlists/{watchlist_id}/tag",
        Resource.watchlists,
        [Action.delete],
    )
    async def remove_watchlist_tag_status(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Disable document tagging for a watchlist
        description: Turn off tagging for watchlist with `watchlist_id`.  This is not retroactive for existing watchlist tags.  Future matches will not trigger event tagging.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_id
            description: ID for watchlist
            required: true
            type: string
        responses:
            204:
                description: Tagging stopped
            400:
                description: Unknown watchlist
        """
        watchlist_id = request.match_info["watchlist_id"]
        org_key = request.match_info["org_key"]

        if not await engine.remove_tag_status(org_key, watchlist_id):
            raise HTTPNotFound(reason=f"Unknown watchlist: {watchlist_id}")

        watchlist_name = await engine.get_watchlist_name(org_key, watchlist_id)

        await call_audit_log(
            request,
            description=f"Disabled tagging for watchlist, '{watchlist_name}' (ID: {watchlist_id})",
        )
        return make_response_json()

    @routes.get(
        "/v3/orgs/{org_key}/reports/{report_id}/ignore",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_report_ignore(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get report ignore status
        description: Get current ignore status for report with `report_id`.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            200:
                description: Returns ignore status
                schema:
                  type: object
                  properties:
                    ignored:
                      type: boolean
        """
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        ignored = await engine.get_report_ignore(org_key, report_id)

        return make_response_json({"ignored": ignored})

    @routes.get(
        "/v3/orgs/{org_key}/reports/{report_id}/iocs/{ioc_id}/ignore",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_report_ioc_ignore(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get IOC ignore status
        description: Get current ignore status for IOC `ioc_id` in report `report_id`
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: path
            name: ioc_id
            description: ID for IOC
            required: true
            type: string
        responses:
            200:
                description: Returns ignore status
                schema:
                  type: object
                  properties:
                    ignored:
                      type: boolean
        """
        report_id = request.match_info["report_id"]
        ioc_id = request.match_info["ioc_id"]
        org_key = request.match_info["org_key"]

        ignored = await engine.get_report_ioc_ignore(org_key, report_id, ioc_id)

        return make_response_json({"ignored": ignored})

    @routes.put(
        "/v3/orgs/{org_key}/reports/{report_id}/ignore",
        Resource.watchlists,
        [Action.update],
    )
    async def add_report_ignore(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Ignore report
        description: Report with `report_id` and all contained IOCs will not match future events for any watchlist.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            200:
                description: Returns ignore status
                schema:
                  type: object
                  properties:
                    ignored:
                      type: boolean
        """
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        # Raise Bad Request exception if request had MDR reports
        if report_id in await engine.get_mdr_reports(org_key):
            raise HTTPBadRequest(
                reason=f"Report Id - {report_id} belongs to mdforeedr, "
                f"therefore can not be ignored."
            )

        await engine.add_ignore_report(org_key, report_id)
        report_title = await engine.get_report_title(report_id)

        await call_audit_log(
            request, description=f"Ignored report, '{report_title}' (ID: {report_id})"
        )
        return make_response_json({"ignored": True})

    @routes.put(
        "/v3/orgs/{org_key}/reports/{report_id}/iocs/{ioc_id}/ignore",
        Resource.watchlists,
        [Action.update],
    )
    async def add_report_ioc_ignore(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Ignore IOC
        description: IOC `ioc_id` for report `report_id` will not match future events for any watchlist.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: path
            name: ioc_id
            description: ID for IOC
            required: true
            type: string
        responses:
            200:
                description: Returns ignore status
                schema:
                  type: object
                  properties:
                    ignored:
                      type: boolean
        """
        report_id = request.match_info["report_id"]
        ioc_id = request.match_info["ioc_id"]
        org_key = request.match_info["org_key"]

        # Raise Bad Request exception if request had MDR reports
        if report_id in await engine.get_mdr_reports(org_key):
            raise HTTPBadRequest(
                reason=f"Report Id - {report_id} belongs to mdforeedr, "
                f"therefore can not be ignored."
            )

        await engine.add_ignore_report_ioc(org_key, report_id, ioc_id)
        report_title = await engine.get_report_title(report_id)

        await call_audit_log(
            request,
            description=f"Ignored IOC, (ID: {ioc_id}) from report, '{report_title}' (ID: {report_id})",
        )
        return make_response_json({"ignored": True})

    @routes.delete(
        "/v3/orgs/{org_key}/reports/{report_id}/ignore",
        Resource.watchlists,
        [Action.delete],
    )
    async def remove_report_ignore(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Re-activate report
        description: Report with `report_id` and all contained IOCs will match future events for all watchlists.  This is not retroactive for events that occured while the report was ignored.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            204:
                description: Report is active
            400:
                description: Unknown report
        """
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        report_title = await engine.get_report_title(report_id)

        await engine.remove_ignore_report(org_key, report_id)

        await call_audit_log(
            request, description=f"Enabled report, '{report_title}' (ID: {report_id})"
        )
        return make_response_json()

    @routes.delete(
        "/v3/orgs/{org_key}/reports/{report_id}/iocs/{ioc_id}/ignore",
        Resource.watchlists,
        [Action.delete],
    )
    async def remove_report_ioc_ignore(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Re-activate IOC
        description: IOC `ioc_id` for report `report_id` and will match future events for all watchlists.  This is not retroactive for events that occured while the IOC was ignored.
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: path
            name: ioc_id
            description: ID for IOC
            required: true
            type: string
        responses:
            204:
                description: Ignore removed - IOC is active
            400:
                description: Unknown report/ioc
        """
        report_id = request.match_info["report_id"]
        ioc_id = request.match_info["ioc_id"]
        org_key = request.match_info["org_key"]

        await engine.remove_ignore_report_ioc(org_key, report_id, ioc_id)
        report_title = await engine.get_report_title(report_id)

        await call_audit_log(
            request,
            description=f"Enabled IOC, (ID: {ioc_id}) from report, '{report_title}' (ID: {report_id})",
        )
        return make_response_json()

    @routes.get(
        "/v3/orgs/{org_key}/reports/severity",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_report_severity_overrides(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get custom report severities
        description: Return all custom report severities.  Custom report severities effect all watchlists.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
        responses:
            200:
                description: Returns list of report severities
                schema:
                  properties:
                    results:
                      type: array
                      items:
                        $ref: '#/definitions/ReportSeverity'
        """
        org_key = request.match_info["org_key"]

        return make_response_json(
            {"results": await engine.get_report_severity_overrides(org_key)}
        )

    @routes.get(
        "/v3/orgs/{org_key}/reports/{report_id}/severity",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_report_severity_override(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get custom severity for report
        description: Return custom severity for `report_id`.  This will return 404 error if custom severity doesn't exist.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            200:
                description: Returns severity.  (null if not set)
                schema:
                  $ref: '#/definitions/ReportSeverity'
            404:
                description: No override for report
        """
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        report_severity = await engine.get_report_severity_override(org_key, report_id)
        if not report_severity:
            raise HTTPNotFound(reason=f"No override for report: {report_id}")

        return make_response_json(report_severity)

    @routes.put(
        "/v3/orgs/{org_key}/reports/{report_id}/severity",
        Resource.watchlists,
        [Action.update],
    )
    async def set_report_severity_override(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Set custom report severity
        description: Adjust the severity of report with `report_id`.  This will effect all watchlists.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            200:
                description: Returns severity.
                schema:
                  $ref: '#/definitions/ReportSeverity'
        """
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        params = await request.json(loads=rapidjson.loads)
        severity = int(_get_required_param(params, "severity"))

        await engine.add_report_severity_override(org_key, report_id, severity)

        report_title = await engine.get_report_title(report_id)

        await call_audit_log(
            request,
            description=f"Set severity to {severity} for report, '{report_title}' (ID: {report_id})",
        )
        return make_response_json(ReportSeverity(report_id, severity))

    @routes.delete(
        "/v3/orgs/{org_key}/reports/{report_id}/severity",
        Resource.watchlists,
        [Action.delete],
    )
    async def remove_report_severity_override(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Remove custom report severity
        description: Remove custom severity for report with `report_id`
        produces:
        - application/octet-stream
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            204:
                description: Severity override removed
        """
        report_id = request.match_info["report_id"]
        org_key = request.match_info["org_key"]

        await engine.remove_report_severity_override(org_key, report_id)

        report_title = await engine.get_report_title(report_id)

        await call_audit_log(
            request,
            description=f"Removed severity from report, '{report_title}' (ID: {report_id})",
        )
        return make_response_json()

    @routes.post("/v3/orgs/{org_key}/reports", Resource.watchlists, [Action.create])
    async def create_report(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Create new report
        description: Add a new watchlist report.  This service will generate a unique report id.  This report will be private to the caller.  IOCs will be converted to IOC_V2.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: body
            name: body
            description: Report payload
            required: true
            schema:
              $ref: '#/definitions/Report'
        responses:
            200:
                description: Report created
                schema:
                  $ref: '#/definitions/Report'
            400:
                description: invalid report request.
        """
        report_id = ReportEngine.gen_report_id()

        params = await request.json(loads=rapidjson.loads)

        org_key = request.match_info["org_key"]

        params["id"] = report_id

        report = await engine.add_report(org_key, Report.from_dict(params))

        await call_audit_log(
            request, description=f"Created report {report.title}({report_id})"
        )
        return make_response_json(report)

    @routes.put(
        "/v3/orgs/{org_key}/reports/{report_id}", Resource.watchlists, [Action.update]
    )
    async def update_report(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Update a report.
        description: Update report with `report_id`.  This will replace all fields in the report.  Any fields not provided in the request will be remove from the report.  All IOCs will be converted to IOC_V2.  The report must be owned by the caller.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: body
            name: body
            description: Report payload
            required: true
            schema:
              $ref: '#/definitions/Report'
        responses:
            200:
                description: Report updated
                schema:
                  $ref: '#/definitions/Report'
            400:
                description: invalid report request.
            404:
                description: report id not found
        """
        report_id = request.match_info["report_id"]

        params = await request.json(loads=rapidjson.loads)

        org_key = request.match_info["org_key"]

        params["id"] = report_id

        report = await engine.update_report(org_key, Report.from_dict(params))

        await call_audit_log(
            request, description=f"Updated report, '{report.title}' (ID: {report_id})"
        )
        return make_response_json(report)

    @routes.get(
        "/v3/orgs/{org_key}/reports/{report_id}",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_report(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get report
        description: Retrieve report with `report_id`.  The report must be owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: query
            name: iocs_start
            description: 0-based row index from which to begin listing report's IOCs. Required if iocs_rows is populated
            required: false
            type: int
          - in: query
            name: iocs_rows
            description: the number of IOCs to return. Required if iocs_start is populated
            required: false
            type: int
        responses:
            200:
                description: Report
                schema:
                  $ref: '#/definitions/Report'
            400:
                description: start param less than 0 or rows param less than 1
            404:
                description: report id not found
        """
        report_id = request.match_info["report_id"]

        org_key = request.match_info["org_key"]

        iocs_start_row = (
            request.query["iocs_start"] if "iocs_start" in request.query else None
        )
        iocs_row_count = (
            request.query["iocs_rows"] if "iocs_rows" in request.query else None
        )

        if iocs_start_row is not None and (
            not iocs_start_row.isdigit() or int(iocs_start_row) < 0
        ):
            raise HTTPBadRequest(
                reason="start parameter must be a non-negative whole number"
            )
        if iocs_row_count is not None and (
            not iocs_row_count.isdigit() or int(iocs_row_count) < 1
        ):
            raise HTTPBadRequest(
                reason="rows parameter must be a positive whole number"
            )
        if iocs_start_row is not None and iocs_row_count is None:
            raise HTTPBadRequest(
                reason="IOC start index provided but missing desired row count"
            )
        if iocs_start_row is None and iocs_row_count is not None:
            raise HTTPBadRequest(
                reason="Desired IOC row count provided but missing start index"
            )

        report = await engine.get_report(
            org_key, report_id, iocs_start_row, iocs_row_count
        )

        return make_response_json(report)

    @routes.delete(
        "/v3/orgs/{org_key}/reports/{report_id}", Resource.watchlists, [Action.delete]
    )
    async def remove_report(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Remove report
        description: Remove report with `report_id`.  The report must be owned by the caller.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
        responses:
            204:
                description: Report deleted
            404:
                description: report id not found
        """
        report_id = request.match_info["report_id"]

        org_key = request.match_info["org_key"]

        report_title = await engine.get_report_title(report_id)

        if not await engine.delete_report(org_key, report_id):
            raise HTTPNotFound(reason=f"Unknown report: {report_id}")

        await call_audit_log(
            request, description=f"Deleted report, '{report_title}' (ID: {report_id})"
        )
        return make_response_json()

    @routes.delete(
        "/v3/orgs/{org_key}/reports/{report_id}/iocs/{ioc_ids}",
        Resource.watchlists,
        [Action.delete],
    )
    async def remove_report_iocs(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Remove report ioc
        description: Remove report iocs with `report_id` and `ioc_ids`.  `ioc_ids` can be a single id or comma seperated list.  The report must be owned by the `org_key`.  `report_timestamp` may be passed as query param and will used to update parent report if timestamp is greater than existing otherwise request will fail.  Current time will be used if not provided.
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: path
            name: ioc_ids
            description: comma seperated list of IOC ids
            required: true
            type: string
        responses:
            204:
                description: Report ioc deleted or does not exist
            400:
                description: report_timestamp older than existing timestamp
            404:
                description: report ioc not found
        """
        org_key = request.match_info["org_key"]

        report_id = request.match_info["report_id"]
        ioc_ids = request.match_info["ioc_ids"].split(",")

        report_timestamp = datetime_from_sec_timestamp(
            int(request.query.get("report_timestamp", time.time()))
        )

        report_title = await engine.get_report_title(report_id)

        await engine.delete_report_iocs(org_key, report_id, ioc_ids, report_timestamp)

        await call_audit_log(
            request,
            description=f"Deleted IOCs, {', '.join(ioc_ids)} from report, '{report_title}' (ID: {report_id}) (v1)",
        )
        return make_response_json()

    @routes.put(
        "/v3/orgs/{org_key}/reports/{report_id}/iocs",
        Resource.watchlists,
        [Action.update],
    )
    async def add_report_iocs(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Add report ioc
        description: Add iocs to report `report_id`.  The report must be owned by the `org_key`.  `report_timestamp` may be passed as query param and will used to update parent report if timestamp is greater than existing otherwise request will fail.  Current time will be used if not provided.
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: body
            name: body
            description: Watchlist payload
            required: true
            schema:
              properties:
                iocs:
                  type: array
                  items:
                    $ref: '#/definitions/IOC_V2'
        responses:
            204:
                description: Report iocs added
            400:
                description: report_timestamp older than existing timestamp
            404:
                description: report ioc not found
        """
        org_key = request.match_info["org_key"]

        report_id = request.match_info["report_id"]

        report_timestamp = datetime_from_sec_timestamp(
            int(request.query.get("report_timestamp", time.time()))
        )

        params = await request.json(loads=rapidjson.loads)
        iocs = [IOC_V2.from_dict(report) for report in params["iocs"]]

        await engine.add_report_iocs(org_key, report_id, iocs, report_timestamp)

        iocs_ids = []
        for ioc in iocs:
            iocs_ids.append(ioc.id)

        report_title = await engine.get_report_title(report_id)

        await call_audit_log(
            request,
            description=f"Added IOCs, (ID:{', '.join(iocs_ids)}) to report, '{report_title}' (ID: {report_id})",
        )
        return make_response_json()

    @routes.get(
        "/v3/orgs/{org_key}/reports/{report_ids}/ignore/bulk",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_bulk_report_ignore_get(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get ignore status for provided report ids (via querystring)
        description: Get current ignore status for report and embedded iocs in provided list of comma-separated `report_ids`.  `report_ids` can be a single id
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_ids
            description: comma seperated list of report ids
            required: true
            type: string
        responses:
            200:
                description: Return list of ignored states
                schema:
                  type: array
                  items:
                    $ref: '#/definitions/ReportIOCIgnore'
        """
        report_ids = request.match_info["report_ids"].split(",")
        org_key = request.match_info["org_key"]

        results = await engine.get_ignores(org_key, report_ids=report_ids)

        return make_response_json({"results": results})

    @routes.post(
        "/v3/orgs/{org_key}/reports/ignore/bulk", Resource.watchlists, [Action.read]
    )
    async def get_bulk_report_ignore_post(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get ignore status for provided report ids (via POST body)
        description: Get current ignore status for reports and embedded iocs in provided JSON array of report_ids
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: body
            name: report_ids
            description: array of report_ids
            required: true
            type: array
        responses:
            200:
                description: Return list of ignored states
                schema:
                  type: array
                  items:
                    $ref: '#/definitions/ReportIOCIgnore'
        """
        params = await request.json(loads=rapidjson.loads)
        org_key = request.match_info["org_key"]

        results = await engine.get_ignores(org_key, report_ids=params["report_ids"])
        return make_response_json({"results": results})

    @routes.put(
        "/v3/orgs/{org_key}/reports/ignore/bulk", Resource.watchlists, [Action.update]
    )
    async def bulk_add_report_ignore(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Bulk ignore report and iocs
        description: All reports and iocs as defined in the `ReportIOCIgnore` list with ignore=True will not match future events for any watchlist.  All items with ignore=False will enable matching on future events.  A `ReportIOCIgnore` that does not define an `ioc_id` will effect the entire report (all iocs).
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: report_id
            description: ID for report
            required: true
            type: string
          - in: body
            name: body
            description: Report IOC Ignore List
            required: true
            schema:
              $ref: '#/definitions/ReportIOCIgnoreList'
        responses:
            200:
                description: Returns ignore status array
                schema:
                  $ref: '#/definitions/ReportIOCIgnoreList'
        """
        org_key = request.match_info["org_key"]

        params = await request.json(loads=rapidjson.loads)
        ignores_list = ReportIOCIgnoreList.from_dict(params)

        mdr_reports = await engine.get_mdr_reports(org_key)

        # Raise Bad Request exception if request had MDR reports
        for ignore in ignores_list.ignores:
            if ignore.report_id in mdr_reports:
                raise HTTPBadRequest(
                    reason=f"Report Id - {ignore.report_id} belongs to mdforeedr, "
                    f"therefore can not be ignored."
                )

        ignores = await engine.bulk_update_report_ioc_ignores(
            org_key, ignores_list.ignores
        )

        await call_audit_log(
            request, description=f"Bulk ignored {len(ignores)} reports"
        )
        return make_response_json(ReportIOCIgnoreList(ignores=ignores))

    @routes.get(
        "/v3/orgs/{org_key}/watchlists/{watchlist_ids}/telemetry",
        Resource.watchlists,
        [Action.read],
        allow_head=False,
    )
    async def get_watchlist_telemetry(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get watchlist telemetry
        description: Returns hits and executions for watchlists over the provided intervals.  By default will return telemetry aggregated over the past day.  Include comma seperated list of intervals in minutes as query param `intervals` to aggregate over different ranges, eg intervals=1440,10080,43200
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: path
            name: watchlist_ids
            description: comma seperated list of watchlist ids
            required: true
            type: string
        responses:
            200:
                description: Returns array of telemetry objects for provided intervals
                schema:
                  $ref: '#/definitions/WatchlistTelemetryList'
        """
        org_key = request.match_info["org_key"]

        intervals = [int(i) for i in request.query.get("intervals", "1440").split(",")]
        watchlist_ids = request.match_info["watchlist_ids"].split(",")

        telemetry = await engine.get_watchlist_telemetry(
            org_id=org_key, watchlist_ids=watchlist_ids, intervals=intervals
        )

        return make_response_json(WatchlistTelemetryList(telemetry=telemetry))

    @routes.post(
        "/v3/orgs/{org_key}/watchlists/telemetry", Resource.watchlists, [Action.read]
    )
    async def get_watchlist_telemetry(request: web.Request):
        """
        ---
        tags:
        - v3
        summary: Get watchlist telemetry
        description: Returns hits and executions for watchlists over the provided intervals. By default, will return telemetry aggregated over the past day.  Include comma seperated list of intervals in minutes as query param `intervals` to aggregate over different ranges, e.g. intervals=1440,10080,43200
        consumes:
        - application/json
        produces:
        - application/json
        parameters:
          - in: path
            name: org_key
            description: Organization key
            required: true
            type: string
          - in: body
            name: body
            description: JSON payload with watchlist_ids and intervals array parameters. Parameter intervals is optional and defaults to single 1440 second interval
            required: true
            schema:
              type: object
              required:
              - watchlist_ids
              properties:
                watchlist_ids:
                  type: array
                  items:
                    type: string
                intervals:
                  type: array
                  items:
                    type: int
        responses:
            200:
                description: Returns array of telemetry objects for provided intervals
                schema:
                  $ref: '#/definitions/WatchlistTelemetryList'
        """
        org_key = request.match_info["org_key"]
        params = await request.json(loads=rapidjson.loads)

        watchlist_ids = params["watchlist_ids"]
        intervals = params.get("intervals", [1440])

        telemetry = await engine.get_watchlist_telemetry(
            org_id=org_key, watchlist_ids=watchlist_ids, intervals=intervals
        )

        return make_response_json(WatchlistTelemetryList(telemetry=telemetry))

    return routes
